import argparse
import ipaddress
import subprocess
import sys

import nettools.sos_infrastructure as sos
from nettools._version import __version__


def do_register_addr(args, ipv4_override=None):
    if not ipv4_override:
        try:
            ipv4 = ipaddress.IPv4Address(args.ipv4)
        except ipaddress.AddressValueError:
            print("Invalid IPv4 passed as argument", file=sys.stderr)
            return 1
    else:
        ipv4 = ipv4_override

    try:
        ipv6 = sos.subnets.register_addr(ipv4, args.fqdn, args.dry, args.no_ipv6)
    except sos.IPNotFoundError:
        print("ERROR: The entered IP wasn't found.")
        return 1
    except subprocess.CalledProcessError:
        print("ERROR: Invoking sans failed.")
        return 1

    if args.no_ipv6:
        print("{} {}".format(args.fqdn, ipv4))
    else:
        print("{} {} {}".format(args.fqdn, ipv4, ipv6))


def do_register_net(args):
    subnet = sos.subnets.netlist[args.net_name]  # type: sos.Subnet
    ipv4 = subnet.find_free_ip()
    if not ipv4:
        print("ERROR: Couldn't find a free IP.")
        return 1

    return do_register_addr(args, ipv4)


def register_host_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--dry', action='store_true', help='dry run, don\'t register')
    parser.add_argument('--no-ipv6', action='store_true', help='only register ipv4 address')

    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    commands = parser.add_subparsers(dest='command', help='commands')

    parser_addr = commands.add_parser('addr', help='register given address')
    parser_addr.add_argument('ipv4', type=str, help='ipv4 address')
    parser_addr.add_argument('fqdn', type=str, help='FQDN')

    parser_net = commands.add_parser('net', help='find free address in network and register')
    parser_net.add_argument('net_name', type=str, help='name of network')
    parser_net.add_argument('fqdn', type=str, help='FQDN')

    commands.add_parser('list', help='list available networks')
    parser.parse_args()

    return parser


def cli():
    parser = register_host_parser()
    args = parser.parse_args()

    if args.command == 'addr':
        return do_register_addr(args)

    elif args.command == 'net':
        try:
            result = do_register_net(args)
        except sos.NotAuthenticatedError:
            return 10
        return result

    elif args.command == 'list':
        print(sos.subnets)
        return 0

    else:
        parser.print_usage()
        return 1
